"""
Common analytical queries and transformations of results.

:::{attention}
This module should be considered experimental and subject to change.
:::
"""

import logging
from datetime import timedelta
from typing import Optional

import geopandas as gpd
import h3.api.numpy_int as h3
import pandas as pd
from shapely.geometry import Polygon

from dfi import models, validate
from dfi.connect import Connect
from dfi.get import Get

_logger = logging.getLogger(__name__)


class Analyse:
    """
    Class with analytical methods to build use cases on top of the DFI's queries.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.

    :param dfi_conn: Instance of a Connect with the credentials and namespace of the DFI connection.
    :example:
    Access via the Client class:
    ```python
    from dfi import Client

    dfi = Client(dfi_token, instance_name, namespace, base_url)
    dfi.analyse.records_in_hexagons(uid, resolution, period)
    ```

    Or access only the Analyse class directly:
    ```python
    from dfi.connect import Connect
    from dfi.get import Analyse

    conn = dfi.Connect(api_token, namespace, instance_name, base_url)
    dfi_analyse = dfi.Analyse(conn)

    dfi_analyse.records_in_hexagons(uid, resolution, period)
    ```
    """

    def __init__(self, dfi_conn: Connect) -> None:
        self.conn: Connect = dfi_conn
        self.get = Get(self.conn)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def add_spatiotemporal_hashing(
        self,
        df_records: pd.DataFrame,
        h3_resolution: int,
        time_resolution_min: int,
    ) -> gpd.GeoDataFrame:
        """
        Finds the unique space-time hashes for an entity_id.

        A space-time hash is created for each record from the H3 id and time period

        | entity_id                            | longitude | latitude | timestamp               |
        |--------------------------------------|-----------|----------|-------------------------|
        | a1a2a3a4-b1b2-c1c2-d1d2-d3d4d5d6d7d8 | -0.0769   | 51.5273  | 2022-01-01 12:32:30.000 |


        With H3 resolution 10 and a time period of 5 minutes the following are appended as columns to the dataframe.

        | h3_id           | period_start          | period_end          | geometry                                                                                                                                                                                                                                                                                      |
        |-----------------|---------------------|---------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
        | 8a194ad3478ffff | 2022-01-01 12:30:00 | 2022-01-01 12:35:00 | POLYGON ((-0.0763811054263752 51.52815508753436, -0.0773394908309314 51.52798511308078, -0.0775149002044645 51.52734623441407, -0.0767319401124731 51.526877335658725, -0.0757735765036792 51.52704731096069, -0.0755981511914388 51.527686184169575, -0.0763811054263752 51.52815508753436)) |


        :param df_records: Dataframe with the records we want to hash.
        :param h3_resolution: Uber's H3 h3_resolution. Allowed numbers are > 1 and < 15.
        :param time_resolution_min: Time interval in minutes for timestamp binning.

        :returns: A copy of df_records casted to a GeoDataFrame with 3 extra columns: [hex_id, period_start, period_end] and
        with "geometry" column the geometry of the H3 hexagon where the record appeared.

        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        df_records = dfi.get.records(entities=[entity_id], time_interval=(start_time, end_time))
        dfi.analyse.add_spatiotemporal_hashing(df_records, h3_resolution=11, time_resolution_min=15)
        ```
        """
        validate.df_records(df_records)
        validate.h3_resolution(h3_resolution)

        df_records["hex_id"] = [
            h3.geo_to_h3(lat, lon, h3_resolution) for lat, lon in zip(df_records["latitude"], df_records["longitude"])
        ]
        df_records = df_records.assign(period_start=lambda df: df["timestamp"].round(f"{time_resolution_min}min"))
        df_records = df_records.assign(period_end=lambda df: df.period_start + timedelta(minutes=time_resolution_min))

        return gpd.GeoDataFrame(
            df_records,
            geometry=df_records["hex_id"].apply(lambda idx: Polygon(h3.h3_to_geo_boundary(idx, geo_json=True))),
        )

    def records_for_entity_id_with_spatiotemporal_hashing(
        self,
        entity_id: str,
        h3_resolution: int,
        time_resolution_min: int,
        time_interval: Optional[models.TimeInterval] = None,
        polygon: Optional[models.Polygon] = None,
    ) -> gpd.GeoDataFrame:
        """
        Finds the unique space-time hashes for an entity_id.

        A space-time hash is created for each record from the H3 id and time period

        | entity_id                            | longitude | latitude | timestamp               |
        |--------------------------------------|-----------|----------|-------------------------|
        | a1a2a3a4-b1b2-c1c2-d1d2-d3d4d5d6d7d8 | -0.0769   | 51.5273  | 2022-01-01 12:32:30.000 |

        With H3 resolution 10 and a time period of 5 minutes the following are appended as columns to the dataframe.

        | h3_id           | period_start          | period_end          | geometry                                                                                                                                                                                                                                                                                      |
        |-----------------|---------------------|---------------------|-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
        | 8a194ad3478ffff | 2022-01-01 12:30:00 | 2022-01-01 12:35:00 | POLYGON ((-0.0763811054263752 51.52815508753436, -0.0773394908309314 51.52798511308078, -0.0775149002044645 51.52734623441407, -0.0767319401124731 51.526877335658725, -0.0757735765036792 51.52704731096069, -0.0755981511914388 51.527686184169575, -0.0763811054263752 51.52815508753436)) |


        :param entity_id: unique identifier of a device we want to analyse.
        :param h3_resolution: Uber's H3 h3_resolution. Allowed numbers are > 1 and < 15.
        :param time_resolution_min: Time interval in minutes for timestamp binning.
        :param time_interval: Tuple of time bounds where (lower bound, upper bound).
        :param polygon: List of vertices [[lon1, lat1], [lon2, lat2], ...] or a list of four
                floats representing the bounding box extremes as [lon_min, lat_min, lon_max, lat_max].
                Non valid input will raise an error.

        :returns:
        A dataframe with the history of the loaded records from DFI by the constraints space and time,
        with extra columns with the binning of H3 resolution and time resolution.

        From there you can get the unique hexagons from where the devices are appearing.

        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.
        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        entity="299eb59a-e47e-48c0-9ad5-89a9ce1303f4"
        polygon = [[-0.1169,51.5096], [-0.1184,51.5090], [-0.1167,51.5074], [-0.1153,51.5079], [-0.1169,51.5096]]
        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")

        dfi.analyse.records_for_entity_id_with_spatiotemporal_hashing(
            entity_id=entity,
            h3_resolution=11,
            time_resolution_min=15,
            time_interval=(start_time, end_time),
            polygon=polygon
        )
        ```
        """
        df_records = self.get.records(entities=[entity_id], polygon=polygon, time_interval=time_interval)

        if len(df_records) == 0:
            _logger.debug("No history found for entity %s", entity_id)
            return gpd.GeoDataFrame(
                columns=[
                    "entity_id",
                    "latitude",
                    "longitude",
                    "timestamp",
                    "hex_id",
                    "period_start",
                    "period_end",
                ]
            )

        return self.add_spatiotemporal_hashing(
            df_records,
            h3_resolution=h3_resolution,
            time_resolution_min=time_resolution_min,
        )

    def add_heatmap_aggregation(
        self,
        df_records: pd.DataFrame,
        h3_resolution: int,
        colorscale_column_name: str = "num_records",
        colorscale_log_transform: bool = True,
    ) -> pd.DataFrame:
        """
        Aggregates the records by the H3 hexagons at the given resolution.

        :param df_records: Dataframe with the records we want to aggregate.
        :param h3_resolution: Uber's H3 resolution. Allowed numbers are > 1 and < 15.
        :param colorscale_column_name: (not implemented) Name of the column with the colorscale for visualisation.
        :param colorscale_log_transform: (not implemented) Log transform the colorscale.

        :returns: The given dataframe with extra `hex_id`, `num_records` and `color` columns.

        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        df_records = dfi.get.records(
            entities=[entity_id],
            time_interval=(start_time, end_time)
        )
        dfi.analyse.heatmap_aggregation(df_records, h3_resolution=11)
        ```
        """
        return (
            df_records.assign(
                hex_id=lambda df: [
                    h3.geo_to_h3(lat, lon, resolution=h3_resolution)
                    for lat, lon in zip(df["latitude"], df["longitude"])
                ]
            )
            .pipe(_aggregate_records, "hex_id")
            .assign(hex_id=lambda df: df.hex_id.map(hex).str[2:])
        )

    def records_for_entity_id_with_heatmap_aggregation(
        self,
        entity_id: str,
        h3_resolution: int,
        time_interval: Optional[models.TimeInterval] = None,
        polygon: Optional[models.Polygon] = None,
        colorscale_colum_name: str = "num_records",
        colorscale_log_transform: bool = True,
    ) -> pd.DataFrame:
        """
        Aggregates the records by the H3 hexagons at the given resolution, from the parameters
        to a DFI query (entity_id, polygon, start_time, end_time)

        :param entity_id: unique identifier of a device we want to analyse.
        :param h3_resolution: Uber's H3 h3_resolution. Allowed numbers are > 1 and < 15.
        :param time_interval: Tuple of time bounds where (lower bound, upper bound).
        :param polygon:  List of vertices [[lon1, lat1], [lon2, lat2], ...] or a list of four
            floats representing the bounding box extremes as [lon_min, lat_min, lon_max, lat_max].
            Non valid input will raise an error.
        :param colorscale_col:  Name of the column with the colorscale for visualisation.
        :param colorscale_log_transform:  Log transform the colorscale.

        :returns: Returns the queried records dataframe with extra `hex_id`, `num_records` and `color` columns.
        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.

        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        entity = "299eb59a-e47e-48c0-9ad5-89a9ce1303f4"
        polygon = [[-0.1169,51.5096], [-0.1184,51.5090], [-0.1167,51.5074], [-0.1153,51.5079], [-0.1169,51.5096]]
        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"),
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ"),

        dfi.analyse.records_for_entity_id_with_heatmap_aggregation(
            entity_id=entity,
            h3_resolution=11,
            polygon=polygon,
            time_interval=(start_time,end_time),
        )
        ```
        """
        df_records = self.get.records(entities=[entity_id], polygon=polygon, time_interval=time_interval)

        if len(df_records) == 0:
            _logger.debug("No history found for entity %s", entity_id)
            return gpd.GeoDataFrame(
                columns=[
                    "entity_id",
                    "latitude",
                    "longitude",
                    "timestamp",
                    "hex_id",
                    "period_start",
                    "period end",
                ]
            )

        return self.add_heatmap_aggregation(
            df_records,
            h3_resolution=h3_resolution,
            colorscale_colum_name=colorscale_colum_name,
            colorscale_log_transform=colorscale_log_transform,
        )


def _aggregate_records(df_input: pd.DataFrame, hex_id: str) -> pd.DataFrame:
    return (
        df_input.groupby(hex_id)
        .agg(
            num_records=("entity_id", "count"),
            num_devices=("entity_id", "nunique"),
            first_ping=("timestamp", "min"),
            last_ping=("timestamp", "max"),
        )
        .reset_index()
    )



"""
Data models for type hints.  For argument validation see `dfi.validate`.

- `Polygon` either a list of polygon vertices or bounding box.
  - vertices - polygon vertices should follow the [GeoJSON RFC7946](https://datatracker.ietf.org/doc/html/rfc7946#section-3.1.6)
    - ex: `[[lon_1, lat_1], [lon_2, lat_2], [lon_3, lat_3], [lon_1, lat_1]]`
  - bbox - bounding box should be a 2D bbox and follow the [GeoJSON RFC7946](https://datatracker.ietf.org/doc/html/rfc7946#section-5)
    - ex: `[min_lon, min_lat, max_lon, max_lat]`
- `TimeInterval` - a tuple of datetimes where the first item should be a datetime before the second item.
    - ex: `( datetime(2022, 1, 1, 0, 0,0 ), datetime(2022, 1, 1, 12, 0, 0) )`
"""
from datetime import datetime
from typing import List, NewType, Optional, Tuple, Union

# TODO: The internal type hinting of these types don't show up in the docs and makes them quite opaque.
# Can fix by using actual types (could use shapely.geometry.Polygon and shapely.geometry.box) which come with validation batteries included.
# Once the 3D Geometry library has a python wrapper this will need to use those types to validate.
Polygon = NewType("Polygon", Union[List[float], List[List[float]]])  # type: ignore
TimeInterval = NewType("TimeInterval", Tuple[Optional[datetime], Optional[datetime]])



"""
Provides a simple way to visualize query results on a map.  For instructions see [Kepler.GL](https://kepler.gl/) user guide.

:::{attention}
This module should be considered experimental and subject to change.
:::
"""

import logging
from copy import deepcopy
from typing import List, Optional, Tuple, Union

import geopandas as gpd
import pandas as pd
from keplergl import KeplerGl
from shapely.geometry import Polygon

from dfi import validate
from dfi.connect import Connect
from dfi.models import Polygon as ModelPolygons
from dfi.polygons import Polygons

_logger = logging.getLogger(__name__)


class Show:
    """
    Visualisation methods to build use case on top of the queries from DFI.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.

    :param dfi_conn: Instance of a Connect with the credentials and namespace of the DFI connection.

    :exmaple:
    Access via the Client class:

    ```python
    from dfi import Client

    dfi = Client(dfi_token, instance_name, namespace, base_url)

    dfi.show.map() # creates an empty Kepler map
    ```
    """

    def __init__(self, dfi_conn: Connect) -> None:
        self.conn: Connect = dfi_conn
        self.polygons = Polygons(self.conn)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def map(
        self,
        list_polygons: Optional[List[ModelPolygons]] = None,
        list_polygons_by_names: Optional[List[str]] = None,
        df_records: Optional[pd.DataFrame] = None,
        list_dfs: Optional[Union[gpd.GeoDataFrame, pd.DataFrame]] = None,
        map_height: int = 1200,
        config: Optional[dict] = None,
    ) -> KeplerGl:
        """
        Helper method to crete a kepler map with the given input layers, to reproduce proposed
        examples in the example notebooks.

        :param list_polygons: List of polygons, as list of vertices [[lon1, lat1], [lon2, lat2], ...].
        :param list_polygons_by_names: List of polygon names as they are named in the auxiliary Polygon DB.
        :param df_records: Dataframe of records with record_id, timestamp, latitude, longitude as its columns.
        :param list_dfs: Generic list of any dataframe or geodataframe to be visualised on the map.
        :param map_height: Parameter passed to the KeplerGl instance as it is, overriding the default.
        :param config: A KeplerGL map config.  See [KeplerGL docs](https://github.com/keplergl/kepler.gl/tree/master/bindings/kepler.gl-jupyter#for-jupyter-notebook-and-jupyterlab) for more info.
        :returns: Instance of a KeplerGL map with the given data to visualise.
        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        df = dfi.get.records(
            entities = ["a1a2a3a4-b1b2-c1c2-d1d2-d3d4d5d6d7d8"]
        )

        dfi.show.map(
            df_records=df
        )
        ```

        :::{note}
        For instructions see [Kepler.GL](https://kepler.gl/) user guide.
        :::
        """
        if list_polygons is None:
            list_polygons = []

        list_polygons_as_vertices = []
        for poly in list_polygons:
            # from bounding boxes to vertices
            validate.polygon(poly)
            if isinstance(poly[0], float):
                list_polygons_as_vertices += [from_bbox_to_polygon_list(poly)]
            else:
                list_polygons_as_vertices += [poly]

        dict_polygons = {f"polygon {idx}": poly for idx, poly in enumerate(list_polygons_as_vertices)}

        if list_polygons_by_names is not None:
            dict_polygons.update(
                {poly_name: self.polygons.get_vertices(poly_name) for poly_name in list_polygons_by_names}
            )

        kepler_data = {}

        if len(dict_polygons) > 0:
            kepler_data.update(
                {
                    "polygons": gpd.GeoDataFrame(
                        dict_polygons.keys(),
                        geometry=[Polygon(x) for x in dict_polygons.values()],
                    )
                }
            )

        if df_records is not None:
            validate.df_records(df_records)
            kepler_data.update({"records": df_records.copy()})

        if list_dfs is not None:
            for idx, df in enumerate(list_dfs):
                kepler_data.update({f"df_{idx}": df.copy()})

        if config is None:
            return KeplerGl(data=deepcopy(kepler_data), height=map_height)
        return KeplerGl(data=deepcopy(kepler_data), height=map_height, config=config)


def from_bbox_to_polygon_list(bounding_box: List[float]) -> Tuple[Tuple[float, float]]:
    """Convert a bounding box, passed as a list `[min_lon, min_lat, max_lon, max_lat]` into a tuple of vertices."""
    validate.bounding_box(bounding_box)
    min_lon, min_lat, max_lon, max_lat = bounding_box
    return ((max_lon, max_lat), (max_lon, min_lat), (min_lon, min_lat), (min_lon, max_lat), (max_lon, max_lat))



"""
Storage and retrieval of named polygons in the DFI Web API.
"""

import logging
from typing import List

from dfi import validate
from dfi.connect import Connect

_logger = logging.getLogger(__name__)


class Polygons:
    """
    Class responsible to call the DFI API and access the polygons DB.

    The polygon DB is a dataset of named polygons. It is made available to the user within the same
    namespace of where the DFI instances lives.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.

    :param dfi_conn: Instance of a Connect with the credentials and namespace of the DFI connection.

    :example:
    Access via the Client class:

    ```python
    from dfi import Client

    dfi = Client(dfi_token, instance_name, namespace, base_url)

    dfi_polygons.get_vertices("my polygon name")
    ```

    Or access only the Polygons class directly:
    ```python
    from dfi.connect import Connect
    from dfi.polygons import Polygons

    conn = dfi.Connect(api_token, instance_name, namespace, base_url)

    dfi_polygons = dfi.Polygons(conn)
    dfi_polygons.get_vertices("my polygon name")
    ```
    """

    def __init__(self, dfi_conn: Connect) -> None:
        self.conn: Connect = dfi_conn

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def get_names(self) -> List[dict]:
        """
        Retrieves a list of saved polygons.

        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        dfi.polygons.get_names()
        ```
        """

        with self.conn.api_get("polygons") as response:
            list_polygons = response.json().get("polygons")
            validate.list_polygons_response(list_polygons, response)
            return list_polygons

    def get_vertices(self, polygon_name: str) -> List[List[float]]:
        """
        Get the list of a polygon saved in the polygon database from its name.

        :param polygon_name: Name of the polygon the user wants to retrieve from the polygons database.
        :returns: List of the polygon coordinates.
        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        dfi.polygons.get_names()
        dfi.polygons.get_vertices('my-first-polygon')
        ```
        """

        with self.conn.api_get("polygons/" + polygon_name) as response:
            vertices = response.json().get("vertices")
            validate.vertices_response(vertices, response)
            return vertices



"""
Class composition of all the other classes, to access the
wrappers with syntactic sugar.
"""
from dfi.analyse import Analyse
from dfi.connect import Connect
from dfi.get import Get
from dfi.polygons import Polygons
from dfi.show import Show


class Client:
    """
    Client class gathering all the classes and method in a single one. Facade of dfi.Connect.

    See documentation of dfi.Connect for its input values.

    :example:
    ```python
    from dfi import Client

    dfi = Client(dfi_token, instance_name, namespace, base_url)
    ```
    """

    def __init__(
        self,
        api_token: str,
        instance_name: str,
        namespace: str = None,
        base_url: str = None,
        query_timeout: int = 60,
        progress_bar: bool = False,
    ) -> None:
        self.conn = Connect(
            api_token=api_token,
            instance_name=instance_name,
            namespace=namespace,
            base_url=base_url,
            query_timeout=query_timeout,
            progress_bar=progress_bar,
        )
        self.get = Get(self.conn)
        self.show = Show(self.conn)
        self.analyse = Analyse(self.conn)
        self.polygons = Polygons(self.conn)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""



"""
Module with private validator methods used across classes.

Some will be improved and delegated to the DFI API.

:example:
```python
from dfi import validate

validate.df_records(df_records)
```

"""
import json
import logging
from datetime import datetime
from typing import List, Optional

import pandas as pd
import requests

from dfi import models

_logger = logging.getLogger(__name__)


class DFIResponseError(Exception):
    """Raised when an error propagated back from the HTTP API"""


class DFIResponseWarning(Warning):
    """Raised when some exceptional case is propagated back from the HTTP API"""


class DFIDataFrameColumnsNameError(Exception):
    """Raised when the column names of a dataframe are not as expected"""


class DFIInputValueError(Exception):
    """Raised when the user passes a wrong input value to the DFI query"""


class DFIInputValueOutOfBoundError(Exception):
    """Raised when the user passes a wrong input value to the DFI query"""


def df_records(df_rec: pd.DataFrame) -> None:
    """
    Check the column names are correct.

    :param h3_res: A dataframe of records
    :returns `None`:
    :raises `DFIDataFrameColumnsNameError`: If a column name is not in
        `["entity_id", "latitude", "longitude", "timestamp"]`
    """
    for col_name in ["entity_id", "latitude", "longitude", "timestamp"]:
        if col_name not in df_rec.columns:
            raise DFIDataFrameColumnsNameError(f"Column name {col_name} expected in df_records but not found.")


def h3_resolution(h3_res: int) -> None:
    """
    check the input is within an acceptable range.

    :param `h3_res`: An H3 resolution
    :returns: `None`
    :raises `DFIInputValueOutOfBoundError`:  If not 1<= h3_res <= 15
    """
    if (h3_res < 1) or (h3_res > 15):
        raise DFIInputValueOutOfBoundError(
            f"Resolution is incorrect. It must be between 1 and 15. User passed {h3_res}"
        )


def vertices_response(vert: Optional[models.Polygon], resp: requests.models.Response) -> None:
    """
    :param vert:
    :returns: `None`
    :raises `DFIResponseError`: If there was an error querying the DFI API.

    :::{TODO}
    Display error from API.
    :::
    """
    if vert is None:
        msg = f"Polygon vertices can not be retrieved from the json response: {resp.json()}"
        _logger.error(msg)
        raise DFIResponseError(msg)


def list_polygons_response(vert: Optional[models.Polygon], resp: requests.models.Response) -> None:
    """
    :param vert:
    :returns: `None`
    :raises `DFIResponseError`: If there was an error querying the DFI API.

    :::{TODO}
    Display error from API.
    :::
    """
    if vert is None:
        msg = f"Polygon list can not be retrieved from the json response: {resp.json()}"
        _logger.error(msg)
        raise DFIResponseError(msg)


def response(
    resp: requests.models.Response,
    url: str,
    headers: dict,
    params: dict,
    payload: Optional[dict] = None,
) -> None:
    """
    Log the response of a request with the given parameters. Raise an error if status code is not 20x.

    :param resp: a response object
    :param url: the queried url
    :param headers: request headers
    :param params: request params
    :param payload: request payload

    :returns: `None`

    :raises `DFIResponseError`: If there was an error querying the DFI API.

    :::{TODO}
    Display error from API.
    :::
    """
    # prevent from showing the user token to terminal and logs
    headers = headers.copy()
    headers["X-API-TOKEN"] = "Bearer XXX"

    msg = f"""Response status code {resp.status_code}.
Query URL: {url},
HEADER: {json.dumps(headers, sort_keys=True, indent=4)},
PARAMS: {json.dumps(params, sort_keys=True, indent=4)}
"""
    if payload is not None:
        msg += f"PAYLOAD: {json.dumps(payload, sort_keys=True, indent=4)}"

    if int(resp.status_code / 10) != 20:
        msg += f" Status Code {resp.status_code}. "
        msg += resp.text
        _logger.error(msg)
        raise DFIResponseError(msg)

    _logger.debug(msg)


def time_interval(time_interv: Optional[models.TimeInterval] = None) -> None:
    """
    Validate input datetimes are both given and compatible.

    :param time_interv: - a tuple of start time and end time bounds e.g. `(start_time, end_time)`
    :returns: `None`
    :raises `DFIInputValueError`: If `time_interv` is ill-formed.
    """
    if time_interv is None:
        return
    if len(time_interv) != 2:
        msg = f"Time interval is not an interval with two dates. User passed {time_interv}"
        raise DFIInputValueError(msg)

    start_time, end_time = time_interv

    if start_time is None and end_time is None:
        return
    if (start_time is None and end_time is not None) or (start_time is not None and end_time is None):
        msg = (
            "start_time and end_time must be both initialised or both None. "
            f"User passed start_time={start_time}, end_time={end_time}"
        )
        raise DFIInputValueError(msg)

    if not isinstance(start_time, datetime):
        msg = f"Start time should be of type datetime. User passed {start_time}"
        raise DFIInputValueError(msg)

    if not isinstance(end_time, datetime):
        msg = f"End time should be of type datetime. User passed {end_time}"
        raise DFIInputValueError(msg)

    if not start_time < end_time:
        msg = f"Start time {start_time} happened after than end time {end_time}."
        raise DFIInputValueError(msg)


def entities(input_entities: Optional[List[str]]) -> None:
    """
    Validate a given list of entities is a list of strings.

    :param `input_entities`: a list of entity ids
    :returns: `None`
    :raises `DFIInputValueError`: If `input_entities` is not a list of string or is empty or has duplicates.
    """
    if input_entities is None:
        return
    if not isinstance(input_entities, list):
        raise DFIInputValueError(f"Entities must be a list of strings. Received {input_entities}")
    if len(input_entities) == 0:
        raise DFIInputValueError("Entities must be a list of strings. Received an empty list")
    if len(set(input_entities)) < len(input_entities):
        duplicates_found = set([x for x in input_entities if input_entities.count(x) > 1])
        raise DFIInputValueError(f"Entities list must not contain duplicates. Duplicates found {duplicates_found}")


def vertices(input_vertices: Optional[List[List[float]]]) -> None:
    """
    Check input list of vertices correspond to a polygon.
    It does not check if the polygon is simple.

    :param input_vertices: a list of polygon vertices
    :returns: `None`
    :raises `DFIInputValueError`: If `input_vertices` is ill-formed.
    """
    if input_vertices is None:
        return
    if len(input_vertices) < 3:
        raise DFIInputValueError(f"A polygon can not have less than 3 vertices. User passed {input_vertices}.")
    for vertex in input_vertices:
        if not len(vertex) == 2:
            raise DFIInputValueError(f"Length of each vertex must be 2. User passed {vertex}")
        if not isinstance(vertex[0], float) or not isinstance(vertex[1], float):
            raise DFIInputValueError(
                f"Coordinates must be of type float."
                f" User passed {vertex} of types ({type(vertex[0])}, {type(vertex[1])})"
            )
        lng, lat = vertex
        if not -180 < lng <= 180:
            raise DFIInputValueOutOfBoundError(f"Input longitude {lng} is out of range.")
        if not -90 < lat < 90:
            raise DFIInputValueOutOfBoundError(f"Input  latitude {lat} is out of range.")

    if not input_vertices[0] == input_vertices[-1]:
        raise DFIInputValueError("First and last vertices are expected to be identical points.")


def bounding_box(list_bbox: Optional[List[float]]) -> None:
    """
    Check input list of coordinates correspond to a bounding box, with lon, lat within the range.

    :param list_bbox: a bbox
    :returns: `None`
    :raises `DFIInputValueError`: If `polygon` is ill-formed.
    :raises `DFIInputValueOutOfBoundError`: If not -180.0 < longitude <= 180.0 or not -90 < latitude <= 90.0.
    """
    if list_bbox is None:
        return
    if len(list_bbox) != 4:
        raise DFIInputValueError(f"Input bounding box parameters must be a list of 4 floats. User passed {list_bbox}")
    for value in list_bbox:
        if not isinstance(value, float):
            raise DFIInputValueError(f"Input value {value} of type {type(value)} must be a float.")

    min_lng, min_lat, max_lng, max_lat = list_bbox

    if not -180 < min_lng <= 180:
        raise DFIInputValueOutOfBoundError(f"Input min longitude {min_lng} is out of range.")
    if not -180 < max_lng <= 180:
        raise DFIInputValueOutOfBoundError(f"Input max longitude {max_lng} is out of range.")
    if not -90 < min_lat < 90:
        raise DFIInputValueOutOfBoundError(f"Input min latitude {min_lat} is out of range.")
    if not -90 < max_lat < 90:
        raise DFIInputValueOutOfBoundError(f"Input max latitude {max_lat} is out of range.")


def polygon(poly: Optional[models.Polygon]) -> None:
    """
    Check input list of coordinates correspond to a list of vertices, or a bounding box.

    :param poly: A list of vertices or bbox.
    :returns `None`:
    :raises `DFIInputValueError`: If `polygon` is ill-formed.
    """
    if poly is None:
        return
    if not isinstance(poly, (list, tuple)):
        raise DFIInputValueError(f"Polygon {poly} must be of type list or tuple.")
    if len(poly) == 0:
        raise DFIInputValueError(f"Given polygon {poly} is empty.")
    if not isinstance(poly[0], (list, tuple, float)):
        raise DFIInputValueError(f"Polygon {poly} must be a list of tuples or floats.")
    if isinstance(poly[0], float):
        bounding_box(poly)
    if isinstance(poly[0], (list, tuple)):
        vertices(poly)


def df_hexes_heatmap(df_h3: pd.DataFrame) -> None:
    """
    Check the column names are correct.

    :param df_h3: A dataframe with a `"hex_id"` column.
    :returns: `None`
    :raises `DFIDataFrameColumnsNameError`: If a column name is not in
        `["entity_id", "latitude", "longitude", "timestamp", "hex_id", "period_start", "period end"]`
    """
    for col_name in ["entity_id", "latitude", "longitude", "timestamp", "hex_id", "period_start", "period end"]:
        if col_name not in df_h3.columns:
            raise DFIDataFrameColumnsNameError(f"Column name {col_name} expected in df_records but not found.")


def df_hexes(df_h3: pd.DataFrame) -> None:
    """
    Check the column names are correct.

    :param df_h3: A dataframe with a `"hex_id"` column.
    :returns: `None`
    :raises `DFIDataFrameColumnsNameError`: If a column name is not in
        `["entity_id", "latitude", "longitude", "timestamp", "hex_id"]`
    """
    for col_name in ["entity_id", "latitude", "longitude", "timestamp", "hex_id"]:
        if col_name not in df_h3.columns:
            raise DFIDataFrameColumnsNameError(f"Column name {col_name} expected in df_records but not found.")



"""
Class to connect to the DFI server.
"""
import logging
from typing import Optional

import requests

from dfi import validate

_logger = logging.getLogger(__name__)


class Connect:
    """
    Class instantiating the connectors to the DFI API.

    :param api_token: token provided by generalsystem.com to access the running DFI environments.
    :param instance_name: Name of a DFI API engine instance, living in the specified namespace.
    :param namespace: Name for the collection of instances of DFI engines where to find the specific instance to connect to.
    :param base_url: Base url where the service is located
    :param query_timeout: Time after an unresponsive query will be dropped.
    :param progress_bar: Visualise a progress bar if True (slows down the execution, typically used for demos and tests).

    :example:
    ````python
    dfi_conn = dfi.Connect(api_token, instance_name, namespace, base_url)

    dfi_conn.streaming_headers
    ````
    """

    def __init__(
        self,
        api_token: str,
        instance_name: str,
        namespace: str = None,
        base_url: str = None,
        query_timeout: int = 60,
        progress_bar: bool = False,
    ) -> None:
        self.api_token = api_token
        self.base_url = base_url
        self.query_timeout = query_timeout
        self.streaming_headers = {
            "X-API-TOKEN": f"Bearer {api_token}",
            "accept": "text/event-stream",
        }
        self.synchronous_headers = {
            "X-API-TOKEN": f"Bearer {api_token}",
            "accept": "application/json",
            "content-type": "application/json",
        }
        self.namespace = namespace
        self.instance_name = instance_name
        self.progress_bar = progress_bar
        if namespace is None:
            self.qualified_instance_name = instance_name
        else:
            self.qualified_instance_name = f"{namespace}.{instance_name}"

    # The representation will expose the secret token
    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self.base_url}"

    def __str__(self) -> str:
        return f"""
                API connection instance
                base_url={self.base_url},
                namespace={self.namespace},
                instance_name={self.instance_name}
                """

    # get and post API wrappers
    def api_get(
        self,
        endpoint: str,
        stream: bool = True,
        params: Optional[dict] = None,
    ) -> requests.models.Response:
        """Helper wrapping requests.get method"""
        headers = self.streaming_headers
        url = f"{self.base_url}/{endpoint}"
        _logger.debug(dict(url=url, headers=headers, stream=stream, params=params, timeout=self.query_timeout))
        response = requests.get(
            url,
            headers=headers,
            stream=stream,
            params=params,
            timeout=self.query_timeout,
        )
        validate.response(response, url, headers, params)
        return response

    def api_post(
        self,
        endpoint: str,
        stream: bool = True,
        params: Optional[dict] = None,
        payload: Optional[dict] = None,
    ) -> requests.models.Response:
        """Helper wrapping requests.post method"""
        headers = self.streaming_headers
        url = f"{self.base_url}/{endpoint}"
        response = requests.post(
            url,
            headers=headers,
            json=payload,
            stream=stream,
            params=params,
            timeout=self.query_timeout,
        )
        validate.response(response, url, headers, params)
        return response



"""
Class with DFI getters, wrappers of the DFI python API.
Composition of the class Connection.
"""
import json
import logging
import warnings
from datetime import datetime
from time import sleep
from typing import Any, List, Optional, Tuple, Union

import pandas as pd
import requests
import sseclient
from tqdm import tqdm

from dfi import models, validate
from dfi.connect import Connect

_logger = logging.getLogger(__name__)

NUM_ATTEMPTS = 3


class Get:
    """
    Class responsible to call the HTTP API and submit queries.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.

    :param conn: Instance of a Connect with the credentials and namespace of the DFI connection.
    :example:
    Access via the Client class:

    ```python
    from dfi import Client

    dfi = Client(dfi_token, instance_name, namespace, base_url)
    dfi.get.entities(time_interval=[start_time, end_time])
    ```

    Or access only the Get class directly:

    ```python
    from dfi.connect import Connect
    from dfi.get import Get

    conn = dfi.Connect(api_token, instance_name, namespace, base_url)
    dfi_get = dfi.Get(conn)
    dfi_get.entities(time_interval=[start_time, end_time])
    ```
    """

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def api_version(self) -> str:
        """
        Returns the version of the Data Flow Index API.
        """
        with self.conn.api_get("version", stream=False, params=None) as response:
            return response.text

    def records_count(
        self,
        entities: Optional[List[str]] = None,
        polygon: Optional[models.Polygon] = None,
        time_interval: Optional[models.TimeInterval] = None,
    ) -> int:
        """
        Queries for the number of records within the bounds.

        If all the variables are None it returns the total number of records in the
        database. Start_time and end_time must be both valid datetime, with
        start_time < end_time or both None.

        :param time_interval: Tuple with the lower bound and the upper bound time constraints.
        :param polyogn: List of vertices `[[lon1, lat1], [lon2, lat2], [lon3, lat3], [lon1, lat1]]` or a list of four
            floats representing the bounding box extremes as `[lon_min, lat_min, lon_max, lat_max]`.
            Non-valid input will raise an error.
        :returns: The number of records stored in the DFI engine.
        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.
        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        entities = ["299eb59a-e47e-48c0-9ad5-89a9ce1303f4"]

        dfi.get.records_count(
            time_interval = (start_time, end_time),
            polygon = None,
            entities = entities,
        )
        ```
        """
        validate.time_interval(time_interval)
        validate.polygon(polygon)
        validate.entities(entities)

        params = {"instance": self.conn.qualified_instance_name}

        if polygon is None and entities is None:
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            with self.conn.api_get("count", params=params) as response:
                return self._receive_count(response)

        if polygon is None and entities is not None:
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            sum_records = 0
            for uid in entities:
                with self.conn.api_get(f"entities/{uid}/count", params=params) as response:
                    sum_records += self._receive_count(response)
            return sum_records

        if isinstance(polygon[0], list) or isinstance(polygon[0], tuple):
            payload = {}
            if time_interval is not None:
                payload["startTime"], payload["endTime"] = unpack_and_convert_time_interval(time_interval)
            if entities is not None:
                payload["include"] = entities
            payload["vertices"] = polygon
            with self.conn.api_post("polygon/count", payload=payload, params=params) as response:
                return self._receive_count(response)

        if isinstance(polygon[0], float):
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            if entities is not None:
                params["include"] = entities

            min_lng, min_lat, max_lng, max_lat = polygon
            with self.conn.api_get(
                f"bounding-box/{min_lng}/{min_lat}/{max_lng}/{max_lat}/count", params=params
            ) as response:
                return self._receive_count(response)

    def records(
        self,
        entities: Optional[List[str]] = None,
        polygon: Optional[models.Polygon] = None,
        time_interval: Optional[models.TimeInterval] = None,
        add_payload_as_json: bool = False,
    ) -> pd.DataFrame:
        """
        Get the records of the entities appearing within the given time, space and entity ids constraints.

        List of entities and polygon can not be left both to None.

        Start time and end time passed in the time_interval must be both valid datetime,
        with start_time < end_time or both None.

        :param time_interval: Tuple with the Lower bound and the upper bound time constraints.
        :param polygon: List of vertices [[lon1, lat1], [lon2, lat2], ...] or a list of four
            floats representing the bounding box extremes as [lon_min, lat_min, lon_max, lat_max].
            Non-valid input will raise an error.
        :param entities: List of entity ids. It must be passed as list also for a single element.
        :param add_payload_as_json: If True it parses the payload as a JSON string into the column payload.
        :returns: Dataframe with the records of the entities found in polygon, given the input constraints.
        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.
        :raises `ValueError`: If no filter bound is specified.
        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        entities = ["299eb59a-e47e-48c0-9ad5-89a9ce1303f4"]

        dfi.get.records(
            time_interval = (start_time, end_time),
            polygon = None,
            entities = entities,
            add_payload_as_json=True
        )

        ```
        """

        validate.time_interval(time_interval)
        validate.polygon(polygon)
        validate.entities(entities)

        params = {"instance": self.conn.qualified_instance_name}

        if polygon is None:
            if entities is None:
                raise ValueError("You have to pass a list of entity ids or a polygon, or both.")
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)

            df_entities = pd.DataFrame(columns=["entity_id", "timestamp", "longitude", "latitude", "payload"])
            for uid in entities:
                with self.conn.api_get(f"entities/{uid}/history", params=params) as response:
                    data = self._receive_history(response)
                    _logger.debug("Uid: %s \nHistory length: %i", uid, len(data))
                    data_formatted = []

                    for item in data:
                        if add_payload_as_json:
                            try:
                                payload = json.loads(item["payload"])
                            except Exception as err:
                                payload = {}
                                _logger.debug("Failed to parse payload to JSON: %s for item %s", err, str(item))
                        else:
                            payload = item["payload"]
                        data_formatted.append(
                            [
                                item["id"],
                                datetime.strptime(item["time"], "%Y-%m-%dT%H:%M:%S.%fZ"),
                                item["coordinate"][0],
                                item["coordinate"][1],
                                payload,
                            ]
                        )
                    df_single_entity = pd.DataFrame(
                        data_formatted, columns=["entity_id", "timestamp", "longitude", "latitude", "payload"]
                    )
                    df_entities = pd.concat([df_entities, df_single_entity])
            return df_entities

        payload = {}
        if time_interval is not None:
            payload["startTime"], payload["endTime"] = unpack_and_convert_time_interval(time_interval)

        if entities is not None:
            payload["include"] = entities

        streamed_data = []

        if isinstance(polygon[0], (list, tuple)):
            # polygon passed by VERTICES
            payload["vertices"] = polygon
            with self.conn.api_post("polygon/history", params=params, payload=payload) as response:
                streamed_data = self._receive_history(response)

        if isinstance(polygon[0], float):
            # polygon passed as a BOUNDING BOX
            min_lng, min_lat, max_lng, max_lat = polygon
            with self.conn.api_get(
                f"bounding-box/{min_lng}/{min_lat}/{max_lng}/{max_lat}/history", params=params
            ) as response:
                streamed_data = self._receive_history(response)

        data_formatted = []
        for item in streamed_data:
            if add_payload_as_json:
                try:
                    payload = json.loads(item.get("payload"))
                except Exception as err:
                    payload = {}
                    _logger.error("Failed to parse payload to JSON: %s", err)
            else:
                payload = item.get("payload")
            data_formatted.append(
                [
                    item["id"],
                    datetime.strptime(item["time"], "%Y-%m-%dT%H:%M:%S.%fZ"),
                    item["coordinate"][0],
                    item["coordinate"][1],
                    payload,
                ]
            )
        return pd.DataFrame(data_formatted, columns=["entity_id", "timestamp", "longitude", "latitude", "payload"])

    def entities(
        self,
        polygon: Optional[models.Polygon] = None,
        time_interval: Optional[models.TimeInterval] = None,
    ) -> List[Union[str, int]]:
        """
        Get the list of entity ids within a space and optional time constraint.

        If time constraints are not passed, the whole dataset is returned.
        Start time and end time passed in the time_interval must be both valid datetime, with
        start_time < end_time or both None.

        :param time_interval: Tuple with the Lower bound and the upper bound time constraints.
        :param polygon: List of vertices [[lon1, lat1], [lon2, lat2], ...] or a list of four
            floats representing the bounding box extremes as [lon_min, lat_min, lon_max, lat_max].
            Non valid input will raise an error.
        :returns: List of unique entities in time interval and polygon.
        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.
        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, instance_name, namespace, base_url)

        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        pimlico_tube_station =  [
            [-0.13410870522207574, 51.48932327401289],
            [-0.1355577905032419, 51.48887697878598],
            [-0.1339996342869938, 51.487266398812636],
            [-0.13279985400049554, 51.487508959676234],
            [-0.13230124401115972, 51.4875768764868],
            [-0.1319428680814667, 51.487955268294115],
            [-0.1322389177630896, 51.4883918703234],
            [-0.1322389177630896, 51.48887697878598],
            [-0.13301799587057417, 51.48930386996324],
            [-0.13410870522207574, 51.48932327401289],
        ]

        dfi.get.entities(
            time_intervals = (start_time, end_time),
            polygon = pimlico_tube_station,
        )

        ```
        """

        validate.time_interval(time_interval)
        validate.polygon(polygon)

        params = {"instance": self.conn.qualified_instance_name}
        payload = {}

        if polygon is None:
            if time_interval is not None:
                payload["startTime"], payload["endTime"] = unpack_and_convert_time_interval(time_interval)
            with self.conn.api_get("entities", params=params, stream=True) as response:
                return self._receive_entities(response)

        if isinstance(polygon[0], list) or isinstance(polygon[0], tuple):
            # polygon passed by VERTICES
            if time_interval is not None:
                payload["startTime"], payload["endTime"] = unpack_and_convert_time_interval(time_interval)
            payload["vertices"] = polygon
            with self.conn.api_post("polygon/entities", params=params, payload=payload) as response:
                return self._receive_entities(response)

        if isinstance(polygon[0], float):
            # polygon passed as a BOUNDING BOX
            params = {"instance": self.conn.qualified_instance_name}
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            min_lng, min_lat, max_lng, max_lat = polygon
            with self.conn.api_get(
                f"bounding-box/{min_lng}/{min_lat}/{max_lng}/{max_lat}/entities", params=params, stream=True
            ) as response:
                self._receive_entities(response)

    def _receive_entities(self, response: requests.models.Response) -> List[Any]:
        """
        Helper function to parse clients events as entities and optionally show the progress bar.
        """
        client = sseclient.SSEClient(response)
        results = []
        results_found = False
        previous = 0
        for event in (pbar := tqdm(client.events(), disable=not self.conn.progress_bar)):
            if event.event == "keepAlive":
                continue
            elif event.event == "finish":
                break
            elif event.event == "message":
                results_found = True
                results += [json.loads(event.data)]

                if self.conn.progress_bar:
                    len_results = len(results)
                    if len_results != previous and len_results % 50 == 0:
                        previous = len(results)
                        pbar.set_description(f"Collecting {previous} records")
                        sleep(0.1)  # to avoid Google Colab being overwhelmed

                continue
            elif event.event == "queryError":
                _raise_query_error_event(event)
            else:
                _raise_unexpected_event_found(event)

        if not results_found:
            _raise_message_event_not_reached()
        return results

    def _receive_history(self, response: requests.models.Response) -> List[Any]:
        """
        Helper function to parse clients events as history and optionally show the progress bar.
        """
        client = sseclient.SSEClient(response)

        results = []
        results_found = False
        previous = 0
        for event in (pbar := tqdm(client.events(), disable=not self.conn.progress_bar)):
            if event.event == "keepAlive":
                continue
            elif event.event == "finish":
                break
            elif event.event == "message":
                results_found = True
                results += json.loads(event.data)

                if self.conn.progress_bar:
                    len_results = len(results)
                    if len_results != previous and len_results % 50 == 0:
                        previous = len(results)
                        pbar.set_description(f"Collecting {previous} records")
                        sleep(0.1)  # to avoid Google Colab being overwhelmed

                continue
            elif event.event == "queryError":
                _raise_query_error_event(event)
            else:
                _raise_unexpected_event_found(event)

        if not results_found:
            _raise_message_event_not_reached()
        return results

    def _receive_count(self, response: requests.models.Response) -> int:
        """
        Helper function to parse clients events as counts and optionally show the progress bar.
        """
        client = sseclient.SSEClient(response)

        results = 0
        previous = 0
        results_found = False
        for event in (pbar := tqdm(client.events(), disable=not self.conn.progress_bar)):
            if event.event == "keepAlive":
                continue
            elif event.event == "finish":
                break
            elif event.event == "message":
                results_found = True
                results = json.loads(event.data)

                if self.conn.progress_bar:
                    if results != previous and results % 50 == 0:
                        previous = results
                        pbar.set_description(f"Collecting {previous} records")
                        sleep(0.1)  # to avoid Google Colab being overwhelmed
                continue
            elif event.event == "queryError":
                _raise_query_error_event(event)
            else:
                _raise_unexpected_event_found(event)

        if not results_found:
            _raise_message_event_not_reached()
        return results


def unpack_and_convert_time_interval(time_interval: models.TimeInterval) -> Tuple[str, str]:
    """from time interval to start_time, end_time in isoformat."""
    # The user can give None to one of the two values here.
    # The time interval validation manages the available options.
    start_time, end_time = time_interval
    if start_time is not None:
        start_time = start_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    if end_time is not None:
        end_time = end_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    return start_time, end_time


def _raise_query_error_event(event: sseclient.Event) -> None:
    """helper to raise a QueryError response error"""
    msg = f"event returned QueryError: {event.data}"
    _logger.error(msg)
    raise validate.DFIResponseError(msg)


def _raise_unexpected_event_found(event: sseclient.Event) -> None:
    """helper to raise a an unexpected event was found"""
    msg = f"Unexpected event in bagging area: {event}"
    _logger.error(msg)
    raise validate.DFIResponseError(msg)


def _raise_message_event_not_reached() -> None:
    """helper to raise a warning if the "message" event was not reached."""
    # see DFIS-694
    msg = "DFI provided no 'message' events."
    _logger.warning(msg)
    warnings.warn(msg)
